//
//  Classes.swift
//  CoreDataMacoun
//
//  Created by Benedikt Terhechte on 10/10/15.
//  Copyright © 2015 Benedikt Terhechte. All rights reserved.
//

import CoreData

protocol StructDecodable {
    func decodeInto(object: NSManagedObject, withKey key: String)
}

extension String: StructDecodable {
    func decodeInto(object: NSManagedObject, withKey key: String) {
        object.setValue(self, forKey: key)
    }
}

extension Int: StructDecodable {
    func decodeInto(object: NSManagedObject, withKey key: String) {
        object.setValue(self, forKey: key)
    }
}

extension NSDate: StructDecodable {
    func decodeInto(object: NSManagedObject, withKey key: String) {
        object.setValue(self, forKey: key)
    }
}

enum SerializationError: ErrorType {
    case StructRequired
    case UnknownEntity(name: String)
    case UnsupportedSubType(label: String?)
}

protocol StructDecoder {
    static var EntityName: String { get }
    func toCoreData(context: NSManagedObjectContext) throws -> NSManagedObject
}

extension StructDecoder {
    func toCoreData(context: NSManagedObjectContext) throws -> NSManagedObject {
        
        let entityName = self.dynamicType.EntityName
        
        guard let desc = NSEntityDescription.entityForName(entityName, inManagedObjectContext: context)
            else { throw SerializationError.UnknownEntity(name: "Entity \(entityName) not found in Core Data model") }
        
        let managedObject = NSManagedObject(entity: desc, insertIntoManagedObjectContext: context)
        
        let mirror = Mirror(reflecting: self)
        
        guard mirror.displayStyle == .Struct else { throw SerializationError.StructRequired }
        
        for case let (label?, valueMaybe) in mirror.children {
            if let value = valueMaybe as? StructDecodable {
                value.decodeInto(managedObject, withKey: label)
            } else {
                let valueMirror = Mirror(reflecting: valueMaybe)
                
                switch (valueMirror.displayStyle, valueMirror.children.first?.value) {
                case (.Optional?, nil): // Leerer Optional
                    break
                case (.Optional?, let value as StructDecodable): // Optional mit value
                    value.decodeInto(managedObject, withKey: label)
                case (.Collection?, _?): // Collection mit Inhalt
                    
                    var convertedObjects = [NSManagedObject]()
                    for childObject in valueMirror.children {
                        if case let (_, value as StructDecoder) = childObject {
                            convertedObjects.append(try value.toCoreData(context))
                        }
                    }
                    
                    if convertedObjects.count > 0 {
                        let mutableValue = managedObject.mutableOrderedSetValueForKey(label)
                        mutableValue.addObjectsFromArray(convertedObjects)
                    }
                    
                default: throw SerializationError.UnsupportedSubType(label: label)
                }
            }
        }
        
        return managedObject
    }
}
